@@ -15,6 +15,7 @@
 #include <stdio.h>
 #include <string.h>
 #include <stdlib.h>
+#include <mntent.h>
 
 
 /*
@@ -25,66 +26,102 @@
 
 
 
-static char *
-get_partition(const char *mountpoint)
+static gboolean
+get_device(glibtop* server, const char *mountpoint,
+	   char* device, size_t device_size)
 {
-	FILE *partitions;
-	char *name = NULL;
-	char line[1024];
-	struct stat statb;
+	const struct mntent *mnt;
+	FILE *fp;
+	gboolean found = FALSE;
 
-	if(stat(mountpoint, &statb) == -1)
-		return NULL;
-
-	if((partitions = fopen("/proc/partitions", "r")) == NULL)
-		return NULL;
-
-	while(fgets(line, sizeof line, partitions))
-	{
-		unsigned major, minor;
-		char dev[32];
-
-		if(sscanf(line, "%u %u %*u %31s", &major, &minor, dev) != 3)
-			continue;
-
-		if(MKDEV(major, minor) != statb.st_dev)
-			continue;
-
-		name = g_strdup(dev);
-		break;
+	if (!(fp = setmntent(MOUNTED, "r"))) {
+		glibtop_warn_io_r(server, "Could not open %s", MOUNTED);
+		goto out;
 	}
 
-	fclose(partitions);
-	return name;
+	while ((mnt = getmntent(fp)))
+	{
+		if (!strcmp(mountpoint, mnt->mnt_dir)) {
+			if (!strncmp(mnt->mnt_fsname, "/dev/", 5)) {
+				g_strlcpy(device, mnt->mnt_fsname + 5, device_size);
+				found = TRUE;
+			}
+			break;
+		}
+	}
+
+	endmntent(fp);
+out:
+	return found;
 }
 
 
-static void
-get_sys_path(const char *device, char **stat_path, const char **parse_format)
+/*
+  TRUE if device is like "hda3" and then set prefix to "hda".
+ */
+static gboolean
+is_partition(const char* device, char* prefix, size_t prefix_size)
 {
-	if(g_str_has_prefix(device, "hd") || g_str_has_prefix(device, "sd"))
-	{
-		char *prefix;
-		char *path;
-		size_t offset;
+	g_strlcpy(prefix, device, prefix_size);
 
-		offset = strcspn(device, "0123456789");
+	for ( ; *prefix; prefix++) {
+		if (isdigit(*prefix)) {
+			*prefix = '\0';
+			return TRUE;
+		}
+	}
 
-		prefix = g_strdup(device);
-		prefix [offset] = '\0';
+	return FALSE;
+}
 
-		path = g_strdup_printf("/sys/block/%s/%s/stat",
-				       prefix, device);
 
-		g_free(prefix);
+/*
+  Bug #539360.
+  /sys/.../stat format is partially defined in
+  linux/Documentation/block/stat.txt (looks outdated).  Before linux
+  2.5.25, /sys/block/%s/stat and /sys/block/%s/%s/stat were not the
+  same, but the following commit changed the latter to have the same
+  format and broke compatibility.
 
-		*stat_path = path;
-		*parse_format = "%*llu %llu %*llu %llu";
+  Commit 34e8beac92c27d292938065f8375842d2840767c
+  Author: Jerome Marchand <jmarchan@redhat.com>
+  Date:   Fri Feb 8 11:04:55 2008 +0100
+
+    Enhanced partition statistics: sysfs
+
+    Reports enhanced partition statistics in sysfs.
+
+    Signed-off-by: Jerome Marchand <jmarchan@redhat.com>
+
+    fs/partitions/check.c |   22 +++++++++++++++++++---
+    1 files changed, 19 insertions(+), 3 deletions(-)
+
+ */
+
+static void
+get_sys_path(glibtop* server, const char *device, char **stat_path, const char **parse_format)
+{
+	const char* linux_2_6_25_format = "%*llu %*llu %llu %*llu"
+					  "%*llu %*llu %llu %*llu";
+
+	char prefix[32];
+
+	if (is_partition(device, prefix, sizeof prefix)) {
+
+		*stat_path = g_strdup_printf("/sys/block/%s/%s/stat",
+					     prefix, device);
+		if (server->os_version_code < LINUX_VERSION_CODE(2, 6, 25))
+			*parse_format = "%*llu %llu %*llu %llu";
+		else
+			*parse_format = linux_2_6_25_format;
 	}
 	else
 	{
 		*stat_path = g_strdup_printf("/sys/block/%s/stat", device);
-		*parse_format = "%*llu %*llu %llu %*llu %*llu %*llu %llu";
+		if (server->os_version_code < LINUX_VERSION_CODE(2, 6, 25))
+			*parse_format = "%*llu %*llu %llu %*llu %*llu %*llu %llu";
+		else
+			*parse_format = linux_2_6_25_format;
 	}
 }
 
@@ -92,17 +129,16 @@ get_sys_path(const char *device, char **stat_path, const char **parse_format)
 
 static void linux_2_6_0(glibtop *server, glibtop_fsusage *buf, const char *path)
 {
-	char *device;
 	char *filename;
 	const char *format;
 	int ret;
 	char buffer[BUFSIZ];
+	char device[64];
 
-	device = get_partition(path);
-	if(!device) return;
+	if (!get_device(server, path, device, sizeof device))
+		return;
 
-	get_sys_path(device, &filename, &format);
-	g_free(device);
+	get_sys_path(server, device, &filename, &format);
 
 	ret = try_file_to_buffer(buffer, sizeof buffer, filename);
 
